(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2004-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: uaref.mli,v 1.27 2014/01/10 15:28:04 fclement Exp $ *)

(** {3 Abstract references and abstract flags}

   {b Abstract references}
   
   An abstract reference is a couple of functions that must be used to get and
   set the contents of the abstract reference.
   A default value is assigned at construction time.
   
   Abstract references may have assignment constraints:
    - abstract references with no assignment constraint {e can} be set many
      times (0+);
    - abstract references with single assignment {e can} be set at most once
      (0,1);
    - abstract references with unique assignement {e must} be set exactly once
      (1);
    - abstract references with mandatory assignment {e must} be set at least
      once (1+).
   
   Abstract references may be protected: they cannot be set anymore when the
   cautious flag is on (see {!set_cautious_flag}).
   Of course, it is meaningless to protect an abstract reference with unique
   assignment.
   
   The contents of abstract references may be bounded (see type
   {!bounds_spec}).
   
   The general mean to obtain an abstract reference value is to build the
   abstract reference specification that matches our needs, and then to
   realize this specification using the generic construction function
   {!make_aref_gen}.
   After construction, an abstract reference cannot be modified anymore.
   Abstract reference specifications are defined incrementally by adding the
   desired properties (assignment constraint, protection, bounds).
   Shortcuts are provided to ease the process when no protection nor bounds
   are needed.
   
   {b Abstract references to lists}
   
   The specific case of abstract references to list contents is treated
   separately, since it is convenient to split the assignment function into
   a clear function and an add function.
   Thus, abstract references to lists are triples of functions.
   
   {b Abstract flags}
   
   Abstract flags are dedicated to the specific case of boolean contents.
   An abstract flag is a triple of functions that must be used to get the
   position of the flag and to set or unset the flag.
   The initial position of a flag is off/down/unset (i.e. value [false]).
   An abstract toggle flag is a couple of functions that must be used to get
   and change the position of the flag.
   A default position is assigned at construction time.
   Shortcuts to obtain a couple of embedded abstract flags and a couple of
   antonymic abstract toggle flags are provided.
   
   Abstract references and abstract flags are particularly useful to define
   the actions associated to command line options (see module
   {!Ucommand_option} (from the OCaml user's additional library Ulib)). *)

(** {6 Abstract references} *)

type 'a aref =
  {
   get : unit -> 'a;
   set : 'a -> unit;
  }
;;
type 'a t = 'a aref;;
(** The type of abstract references.
  Field [get] must be used to get the contents of the abstract reference,
  and field [set] must be used to set the contents of the abstract
  reference. *)

type 'a aref_list =
  {
   get_list : unit -> 'a list;
   clear_list : unit -> unit;
   add_item : 'a -> unit;
  }
;;
(** The type of abstract reference to lists.
   Field [get_list] must be used to get the contents of the abstract reference,
   the field [clear_list] must be used to clear the contents of the abstract
   reference, and the field [add_item] must be used to add an element to the
   contents of the abstract reference. *)

val make_aref : 'a -> 'a t;;
(** Construction function for abstract references with no specific property.
   The argument is the default value. *)

val make_aref_list : 'a list -> 'a aref_list;;
(** Construction function for abstract references to lists with no specific
   property.
   The argument is the initial value. *)

type aref_name = string;;
(** The type of names of abstract references.
   It is used inside error messages. *)

val make_aref_single : aref_name -> 'a -> 'a t;;
val make_aref_unique : aref_name -> 'a -> 'a t;;
val make_aref_mandatory : aref_name -> 'a -> 'a t;;
(** Construction functions for abstract references with assignment constraints.
   The first argument is the name of the abstract reference.
   The second argument is the default value for abstract references with
   single assignment, and the initial value for abstract references with
   unique or mandatory assignment.
   
   Any forbidden assignment or invalid fetch raises [Failure] with a message
   containing the name of the abstract reference. *)

val make_aref_list_min_max :
  aref_name -> 'a list -> Unatural.t -> Unatural.t -> 'a aref_list
;;
(** Construction function for abstract references to lists of given length.
   [make_aref_list_min_max name init min max] returns an abstract reference to
   a list with name [name], initial contents [init], and desired length of
   contents is between [min] and [max] (inclusive).

   Raise [Failure] with a message containing the name of the abstract
   reference when the desired minimum length [min] is nonpositive, or when
   [max] is greater than [min]. *)

val make_aref_list_n : aref_name -> 'a list -> Unatural.t -> 'a aref_list;;
(** Construction function for abstract references to lists of given length.
   [make_aref_list_n name init n] returns an abstract reference to a list
   with name [name], initial contents [init], and desired length of contents
   is [n].

   Raise [Failure] with a message containing the name of the abstract
   reference when the desired length [n] is nonpositive. *)

(** {3 Abstract reference specifications} *)

(** Abstract reference specifications are used to specify the properties of an
   abstract reference (name, assignment constraint, protection, and bounds).
   Dedicated functions are provided to add those properties.
   Properties can be added only once. *)

type ('a, 'b) aref_spec;;
(** The type of abstract reference specifications.
   Type argument ['a] refers to the type of the default value,
   and type argument ['b] refers to the type of bounds specification
   (see {!bounds_spec}). *)

val aref_spec : 'a -> ('a, 'b) aref_spec;;
(** Abstract reference specification construction function.
   The argument is the default value.
   No particular property is specified.
   To incrementally specify properties (name, assignment constraint,
   protection, and bounds), use the dedicated [with_*] functions below.
   See {!make_aref_gen} for usage examples. *)

val make_aref_gen : ('a, 'a) aref_spec -> 'a t;;
(** Abstract reference construction function from its abstract reference
   specification.
   
   {b Examples}
   
   [make_aref_gen (aref_spec 0)] returns an abstract reference with integer
   contents, and with default value zero;
   it is equivalent to [make_aref 0].
   
   [make_aref_gen (with_name "epsilon" (with_unique (aref_spec 0.)))] returns
   an abstract reference with float contents, with reference name ["epsilon"],
   and that must be set exactly once (see {!with_unique_assignment});
   it is equivalent to [make_aref_unique "epsilon" 0.].
   
   [make_aref_gen
    (with_name "log_file"
      (with_protection (aref_spec "/tmp/foo")))]
   returns an abstract reference with string contents, with default value
   ["/tmp/foo"], with reference name ["log_file"], and that cannot be set when
   the cautious flag is on (see {!with_protection}).
   
   [make_aref_gen (with_int_bounds (Some 0) (Some 9) (aref_spec 0))] returns
   an abstract reference with integer contents, with empty reference name,
   with default value [0], and that can only be set with a value in the range
   \[0..9\] (see {!with_int_bounds}).
   
   [make_aref_gen
      (with_int_bounds (Some 1) None
        (with_name "number_of_worker"
          (with_mandatory (aref_spec 2))))]
   returns an abstract reference with integer contents, with reference name
   ["number_of_worker"], that must be assigned at least once
   (see {!with_mandatory_assignment}), and that can only be set with a value
   greater than zero (see {!with_int_bounds}). *)

val make_aref_list_gen : ('a list, 'a) aref_spec -> 'a aref_list;;
(** Construction function for abstract references to lists from their abstract
   reference specification.
   
   {b Examples}
   
   [make_aref_list_gen (aref_spec [ 0; ])] returns an abstract reference with
   integer list contents, and with contents initialized to [[ 0; ]];
   it is equivalent to [make_aref_list [ 0; ]].
   
   [make_aref_list_gen
      (with_name "numbers"
        (with_assignment_constraint (Some 3) (Some 3)
          (aref_spec [ "one"; "two"; "three"; ])))]
   returns an abstract reference with string list contents, with reference
   name ["numbers"], and that can be fetched only when the list contains three
   elements;
   it is equivalent to
   [make_aref_list_n "numbers" [ "one"; "two"; "three"; ] 3].
   
   [make_aref_list_gen
      (with_name "weekdays"
        (with_assignment_constraint None (Some 7)
          (aref_spec [ "sunday"; "wednesday"; ])))]
   returns an abstract reference with string list contents, with reference
   name ["weekdays"], and that can be fetched only when the list contains
   up to seven elements. *)

(** {3 Abstract references with assignment constraints} *)

(** Any attempt to assign or modify the contents of an abstract reference more
   times than the specified maximum, or to fetch the contents of an abstract
   reference before it has been assigned or modified at least the specified
   minimum of times raises [Failure] with a message containing the name of the
   abstract reference. *)

val with_name : aref_name -> ('a, 'b) aref_spec -> ('a, 'b) aref_spec;;
(** Incremental addition of the name property to abstract reference
   specifications.
   See {!make_aref_gen} for usage examples.
   
   Raise [Failure] when the name property is added twice. *)

val with_assignment_constraint :
  Unatural.t option -> Unatural.t option ->
  ('a, 'b) aref_spec -> ('a, 'b) aref_spec
;;
(** Incremental addition of the assignment constraint property to abstract
   reference specifications.
   [with_assignment_constraint min max aref_spec] returns the same abstract
   reference specification [aref_spec] with the additional constraint on the
   possible numbers of assignments specified by the bounds [min] (value [None]
   means zero) and [max] (value [None] means plus infinity).
   See {!make_aref_list_gen} for usage examples.
   
   Raise [Failure] when the assignment constraint property is added twice,
   when the specified minimum is negative, or is greater than the specified
   maximum. *)

val with_single_assignment : ('a, 'b) aref_spec -> ('a, 'b) aref_spec;;
val with_unique_assignment : ('a, 'b) aref_spec -> ('a, 'b) aref_spec;;
val with_mandatory_assignment : ('a, 'b) aref_spec -> ('a, 'b) aref_spec;;
val with_n_assignments :
  Unatural.t -> ('a, 'b) aref_spec -> ('a, 'b) aref_spec
;;
(** Shortcuts:
   - [with_single_assignment] is equivalent to
     [with_assignment_constraint None (Some Unatural._1_)],
     it is used to specify abstract references with single assignment;
   - [with_unique_assignment] is equivalent to
     [with_assignment_constraint (Some Unatural._1_) (Some Unatural._1_)],
     it is used to specify abstract references with unique assignment;
   - [with_mandatory_assignment] is equivalent to
     [with_assignment_constraint (Some Unatural._1_) None],
     it is used to specify abstract references with mandatory assignment;
   - [with_n_assignments n] is equivalent to
     [with_assignment_constraint (Some n) (Some n)],
     it is used to specify abstract references with fixed number of
     assignements.
   
   See {!make_aref_gen} and {!make_aref_list_gen} for usage examples.
   
   Raise [Failure] when the assignment constraint property is added twice. *)

(** {3 Abstract references with protected contents} *)

(** Any attempt to assign a protected abstract reference when the cautious
   flag is on raises [Failure] with a message containing the name of the
   abstract reference. *)

val set_cautious_flag : unit -> unit;;
val unset_cautious_flag : unit -> unit;;
(** To set or unset the cautious flag for driving the assignment of protected
   abstract references.
   Default is on. *)

val with_protection : ('a, 'b) aref_spec -> ('a, 'b) aref_spec;;
(** Incremental addition of the protection property to abstract reference
   specifications.
   See {!make_aref_gen} for usage examples.
   
   Raise [Failure] when the protection property is added twice, or added to an
   abstract reference specification with unique assignment property. *)

(** {3 Abstract references with bounded contents} *)

(** Any attempt to assign an out-of-bound value to the contents of a bounded
   abstract reference raises [Failure] with a message containing the name of
   the abstract reference and the violated bound. *)

type 'b bounds_spec =
  {
   min : 'b option;
   max : 'b option;
   compare : 'b -> 'b -> int;
   string_of : 'b -> string;
  }
;;
(** Bounds property for abstract references specifications.
   Fields [min] and [max] are options to specify the minimum and the maximum
   (the value [None] means no bound).
   The [compare] field is a comparison function that must return 0
   if its arguments compare as equal, a positive integer if the first is
   greater, and a negative integer if the first is smaller (see [Array.sort]);
   [Pervasives.compare] is a suitable candidate.
   The field [string_of] is used to print error messages. *)

val with_bounds : 'b bounds_spec -> ('a, 'b) aref_spec -> ('a, 'b) aref_spec;;
(** Incremental addition of the bounds property to abstract reference
   specifications.
   See {!make_aref_gen} for usage examples.
   
   Raise [Failure] when the bounds property is added twice, or if the
   specified minimum bound is greater than the maximum bound according to the
   comparison function. *)

val with_int_bounds :
  int option -> int option -> ('a, int) aref_spec -> ('a, int) aref_spec
;;
(** Shortcut to add the bounds property with integer bounds.
   The first argument specifies the minimum ([None] means minus infinity), and
   the second argument specifies the maximum ([None] means plus infinity).
   The comparison function is set to [Pervasives.compare] and the field
   [string_of] is set to [Pervasives.string_of_int]. *)

(** {6 Abstract flags} *)

type flag =
  {
   get_flag : unit -> bool;
   set_flag : unit -> unit;
   unset_flag : unit -> unit;
  }
;;
(** The type of abstract flags.
   Field [get_flag] must be used to get the position of the abstract flag,
   and fields [set_flag], or [unset_flag], must be used to set, or unset,
   the abstract flag. *)

val make_flag : unit -> flag;;
(** Abstract flag construction function.
   Return an abstract flag with initial position off/down/unset (i.e. value
   [false]).
   The [set_flag] function assigns the abstract flag to the position
   on/up/set (i.e. value [true]), and the [unset_flag] function assigns it
   back to its initial position. *)

val make_embedded_flags : unit -> flag * flag;;
(** [make_embedded_flags ()] returns a pair of abstract flags [(fl1, fl2)]
   both with initial position off/down/unset (i.e. value [false]).
   Setting [fl2] also sets [fl1]; unsetting [fl1] also unsets [fl2].
   
   Typically, this is useful to implement reporting and debugging option flags
   since the debugging information contain the reporting messages. *)

type tflag =
  {
   get_tflag : unit -> bool;
   toggle_tflag : unit -> unit;
  }
;;
(** The type of abstract toggle flags.
   Field [get_tflag] must be used to get the position of the abstract toggle
   flag, and field [toggle_tflag] must be used to change the position of the
   abstract toggle flag. *)

val make_toggle_flag : bool -> tflag;;
(** Abstract toggle flag construction function.
   Return an abstract toggle flag with initial position [b] ([false] means
   off/down/unset, and [true] means on/up/set).
   The [toggle_tflag] function changes the abstract toggle flag to its other
   position. *)

val make_antonymic_toggle_flags : bool -> tflag * tflag;;
(** [make_antonymic_toggle_flags b] returns a pair of abstract toggle flags
   [(tfl1, tfl2)]; [tfl1] has initial position [b] (see {!make_toggle_flag}),
   and the positions of [tfl1] and [tfl2] are always antonymic. *)
