(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2004-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: ucommand_option.ml,v 1.53 2014/01/10 15:28:04 fclement Exp $ *)

let co_invalid_arg = Upervasives.invalid_arg "Ucommand_option";;
let co_failwith = Upervasives.failwith "Ucommand_option";;

let hashtbl_list_add t k x =
  if not (Hashtbl.mem t k) then Hashtbl.add t k [ x; ] else
  let l = Hashtbl.find t k in
  Hashtbl.replace t k (x :: l)
;;

(* Defining command options. *)

type regular_command_option =
  {
   key : Arg.key;
   spec : Arg.spec;
   doc : Arg.doc;
  }
;;

let triple_of_regular_command_option { key = k; spec = s; doc = d; } =
  (k, s, d)
;;

type command_option =
  {
   adv_key : Arg.key;
   adv_gnu_key : Arg.key option;
   adv_short_key : Arg.key option;
   adv_key_aliases : Arg.key list;
   adv_spec : Arg.spec;
   adv_doc : Arg.doc;
  }
;;

type t = command_option;;

let of_regular { key = key; spec = spec; doc = doc; } =
  {
   adv_key = key;
   adv_gnu_key = None;
   adv_short_key = None;
   adv_key_aliases = [];
   adv_spec = spec;
   adv_doc = doc;
  }
;;

let with_gnu_key key cmd_opt =
  { cmd_opt with adv_gnu_key = Some key; }
;;

let with_short_key key cmd_opt =
  { cmd_opt with adv_short_key = Some key; }
;;

let with_key_aliases keys cmd_opt =
  { cmd_opt with adv_key_aliases = keys; }
;;

let with_auto_gnu_key ({ adv_key = key; _ } as cmd_opt) =
  with_gnu_key ("-" ^ key) cmd_opt
;;

let with_auto_short_key ({ adv_key = key; _ } as cmd_opt) =
  with_short_key (Ustring.short_key_of_long_key key) cmd_opt
;;

(* Managing command options. *)

let ({ Uaref.get_list = get_usr_command_options;
      clear_list = clear_usr_command_options;
      add_item = add_usr_command_option; } as usr_co_arl) =
  Uaref.make_aref_list []
and ({ Uaref.get_list = get_lib_command_options;
      clear_list = clear_lib_command_options;
      add_item = add_lib_command_option; } as lib_co_arl) =
  Uaref.make_aref_list []
;;

let check_unit_spec = function
  | Arg.Unit spec -> spec
  | Arg.Bool _
  | Arg.Set _
  | Arg.Clear _
  | Arg.String _
  | Arg.Set_string _
  | Arg.Int _
  | Arg.Set_int _
  | Arg.Float _
  | Arg.Set_float _
  | Arg.Tuple _
  | Arg.Symbol _
  | Arg.Rest _ -> co_invalid_arg "" "not a Arg.Unit Arg.spec"
;;

let make_co_spec_aref_lists () = 
  let ({ Uaref.clear_list = clear_cos;
         add_item = add_co; _; } as co_arl) = Uaref.make_aref_list [] in
  let { Uaref.get_list = get_specs;
        clear_list = clear_specs;
        add_item = add_spec; } = Uaref.make_aref_list [] in
  let clear_cos () = clear_cos (); clear_specs (); in
  let add_co co = add_co co; add_spec (check_unit_spec co.adv_spec); in
  ({ co_arl with Uaref.clear_list = clear_cos; add_item = add_co; }, get_specs)
;;

let (({ Uaref.get_list = get_report_usr_command_options;
        clear_list = clear_report_usr_command_options;
        add_item = add_report_usr_command_option; } as report_usr_co_arl),
     get_report_usr_specs) = make_co_spec_aref_lists ()
and (({ Uaref.get_list = get_report_lib_command_options;
        clear_list = clear_report_lib_command_options;
        add_item = add_report_lib_command_option; } as report_lib_co_arl),
     get_report_lib_specs) = make_co_spec_aref_lists ()
;;

let get_report_command_options () =
  Ulist.concat
    [ get_report_usr_command_options ();
      get_report_lib_command_options (); ]
and get_report_specs () =
  Ulist.concat
    [ get_report_usr_specs ();
      get_report_lib_specs (); ]
;;

let (({ Uaref.get_list = get_debug_usr_command_options;
        clear_list = clear_debug_usr_command_options;
        add_item = add_debug_usr_command_option; } as debug_usr_co_arl),
     get_debug_usr_specs) = make_co_spec_aref_lists ()
and (({ Uaref.get_list = get_debug_lib_command_options;
        clear_list = clear_debug_lib_command_options;
        add_item = add_debug_lib_command_option; } as debug_lib_co_arl),
     get_debug_lib_specs) = make_co_spec_aref_lists ()
;;

let get_debug_command_options () =
  Ulist.concat
    [ get_debug_usr_command_options ();
      get_debug_lib_command_options (); ]
and get_debug_specs () =
  Ulist.concat
    [ get_debug_usr_specs ();
      get_debug_lib_specs (); ]
;;

let { Uaref.get_list = get_service_command_options;
      clear_list = _clear_service_command_options;
      add_item = add_service_command_option; } =
  Uaref.make_aref_list_min_max
    "service_command_options" [] Unatural._1_ Unatural._3_
;;

let user_add = add_usr_command_option
and library_add = add_lib_command_option
and report_user_add = add_report_usr_command_option
and report_library_add = add_report_lib_command_option
and debug_user_add = add_debug_usr_command_option
and debug_library_add = add_debug_lib_command_option
;;

let user_add_all = Ulist.iter user_add
and library_add_all = Ulist.iter library_add
and report_user_add_all = Ulist.iter report_user_add
and report_library_add_all = Ulist.iter report_library_add
and debug_user_add_all = Ulist.iter debug_user_add
and debug_library_add_all = Ulist.iter debug_library_add
;;

let user_clear = clear_usr_command_options
and library_clear = clear_lib_command_options
and report_user_clear = clear_report_usr_command_options
and report_library_clear = clear_report_lib_command_options
and debug_user_clear = clear_debug_usr_command_options
and debug_library_clear = clear_debug_lib_command_options
;;

let clear_all () =
  user_clear ();
  library_clear ();
  report_user_clear ();
  report_library_clear ();
  debug_user_clear ();
  debug_library_clear ();
;;

let clear_key co_arl key =
  let col = co_arl.Uaref.get_list () in
  let col = Ulist.filter (fun co -> co.adv_key <> key) col in
  co_arl.Uaref.clear_list ();
  Ulist.iter co_arl.Uaref.add_item col;
;;

let user_clear_key = clear_key usr_co_arl
and library_clear_key = clear_key lib_co_arl
and report_user_clear_key = clear_key report_usr_co_arl
and report_library_clear_key = clear_key report_lib_co_arl
and debug_user_clear_key = clear_key debug_usr_co_arl
and debug_library_clear_key = clear_key debug_lib_co_arl
;;

(* Parsing command options. *)

type command_line = string array;;

let { Uaref.get = get_argv; set = set_argv; } = Uaref.make_aref Sys.argv;;

let (default_anon_fun : string -> unit) = ignore;;
let { Uaref.get = get_anon_fun; set = set_anon_fun; } =
  Uaref.make_aref_single "Anonymous function" default_anon_fun
;;

let default_usage_msg =
  Printf.sprintf
    "\n\
     Usage: %s [options]\n\
     \n\
     Options are:"
    Sys.argv.(0)
;;
let { Uaref.get = get_usage_msg; set = set_usage_msg; } =
  Uaref.make_aref_single "Usage message" default_usage_msg
;;

let add_alias_doc key len i accu alias =
  let alias_plural_mark = if len <= 1 then "" else "es" in
  let prefix =
    if i > 0 then "" else
    Printf.sprintf "\n\tAlias%s for %s: " alias_plural_mark key in
  let suffix = if i >= pred len then "." else ", " in
  Printf.sprintf "%s%s%s%s" accu prefix alias suffix
;;

let add_aliases_doc key doc aliases =
  let len = Unatural.to_int (Ulist.length aliases) in
  Ulist.fold_left_i (add_alias_doc key len) doc aliases
;;

let expand_command_option adv_cmd_opt =
  let rec loop col key_aliases = function
    | { adv_key = key;
        adv_gnu_key = None;
        adv_short_key = None;
        adv_key_aliases = [];
        adv_spec = spec;
        adv_doc = doc; } ->
          let doc =
            let doc = add_aliases_doc key doc (Ulist.rev key_aliases) in
            if doc = "" then doc else Printf.sprintf "%s\n" doc in
          { key = key; spec = spec; doc = doc; } :: col
    | { adv_gnu_key = Some gnu_key;
        adv_key_aliases = l; _ } as adv_cmd_opt ->
          loop col key_aliases
            { adv_cmd_opt with
              adv_gnu_key = None; adv_key_aliases = gnu_key :: l; }
    | { adv_short_key = Some short_key;
        adv_key_aliases = l; _ } as adv_cmd_opt ->
          loop col key_aliases
            { adv_cmd_opt with
              adv_short_key = None; adv_key_aliases = short_key :: l; }
    | { adv_key_aliases = alias_key :: l;
        adv_spec = spec; _ } as adv_cmd_opt ->
          loop
            ({ key = alias_key; spec = spec; doc = ""; } :: col)
            (alias_key :: key_aliases)
            { adv_cmd_opt with adv_key_aliases = l; }
  in
  loop [] [] adv_cmd_opt
;;

type option_class =
  | User    | Report_user    | Debug_user
  | Library | Report_library | Debug_library
  | Service
;;

let display_order =
  [ User;    Report_user;    Debug_user;
    Library; Report_library; Debug_library;
    Service; ]
;;
let all_option_classes_but_service =
  Ulist.find_all (fun oc -> oc <> Service) display_order
;;

let get_class_command_options = function
  | User -> get_usr_command_options ()
  | Report_user -> get_report_usr_command_options ()
  | Debug_user -> get_debug_usr_command_options ()
  | Library -> get_lib_command_options ()
  | Report_library -> get_report_lib_command_options ()
  | Debug_library -> get_debug_lib_command_options ()
  | Service -> get_service_command_options ()
;;

let gen_get_all_command_options ocs =
  let f accu oc = get_class_command_options oc :: accu in
  let coll = Ulist.fold_left f [] (Ulist.rev ocs) in
  Ulist.concat coll
;;

let get_all_command_options () = gen_get_all_command_options display_order;;
let get_all_command_options_but_service () =
  gen_get_all_command_options all_option_classes_but_service
;;

let lauch_spec spec = spec ();;
let lauch_specs specs = Ulist.iter lauch_spec specs;;
(* Launch first specified library spec first. *)
let launch_report_specs () = lauch_specs (Ulist.rev (get_report_specs ()))
and launch_debug_specs () = lauch_specs (Ulist.rev (get_debug_specs ()))
;;

let add_rd kind launch_specs = function
  | [] -> ()
  | _ :: _ ->
      let adv_cmd_opt =
        of_regular
          {
           key = Printf.sprintf "-%s" kind;
           spec = Arg.Unit launch_specs;
           doc =
             Printf.sprintf
               "\n\tTurn all declared %s modes on. Default is off."
               kind;
          } in
      add_service_command_option
        (with_auto_short_key (with_auto_gnu_key adv_cmd_opt))
;;

let add_report_and_debug_options () =
  add_rd "report" launch_report_specs (get_report_command_options());
  add_rd "debug" launch_debug_specs (get_debug_command_options ());
;;

let rec add_help_option () =
  let adv_cmd_opt =
    of_regular
      {
       key = "-help";
       spec =
         Arg.Unit
           (fun () -> (* Recovery of speclist must be delayed. *)
             let speclist = get_speclist () in
             let usage_msg = get_usage_msg () in
             raise (Arg.Help (Arg.usage_string speclist usage_msg)));
       doc = "\n\tDisplay this list of options.";
      } in
  add_service_command_option
    (with_auto_short_key (with_auto_gnu_key adv_cmd_opt));
and add_all_service_command_options_once =
  let todo = ref true in
  (fun () ->
    if !todo then
      begin
        todo := false; (* We wont loop forever. *)
        add_report_and_debug_options ();
        add_help_option ();
      end)
and get_speclist () =
  add_all_service_command_options_once ();
  let adv_cmd_opts = get_all_command_options () in
  let cmd_opts = Ulist.flat_map expand_command_option adv_cmd_opts in
  Ulist.map triple_of_regular_command_option cmd_opts
;;

let parse_argv argv =
  let speclist = get_speclist () in
  let anon_fun = get_anon_fun () in
  let usage_msg = get_usage_msg () in
  Arg.current := 0;
  try Arg.parse_argv argv speclist anon_fun usage_msg with
  | Arg.Bad msg -> Printf.eprintf "%s" msg; exit 2;
  | Arg.Help msg -> Printf.printf "%s" msg; exit 0;
;;

let parse () =
  let argv = get_argv () in
  parse_argv argv;
  set_argv [||]
;;

(* Phased parsing of command line options. *)

type phase_name = Phase of string;;

let phase_command_options = Hashtbl.create 3;;

let phase_add phase_name key =
  let cmd_opts = get_all_command_options_but_service () in
  let cmd_opt =
    try Ulist.find (fun cmd_opt -> cmd_opt.adv_key = key) cmd_opts with
    | Not_found ->
        co_failwith "phase_add"
          (Printf.sprintf "key %S is not already recorded" key) in
  hashtbl_list_add phase_command_options phase_name cmd_opt
;;

let phase_add_all phase_name = Ulist.iter (phase_add phase_name);;

let get_keys speclist = Ulist.map (fun (key, _, _) -> key) speclist;;

let find_spec key speclist =
  let (_, spec, _) = Ulist.find (fun (k, _, _) -> k = key) speclist in
  spec
;;

let rec append_args key (accu, argl) = function
  | Arg.Unit _
  | Arg.Clear _ -> (key :: accu, argl)
  | Arg.Bool _
  | Arg.Set _
  | Arg.String _
  | Arg.Set_string _
  | Arg.Int _
  | Arg.Set_int _
  | Arg.Float _
  | Arg.Set_float _
  | Arg.Symbol _ ->
    begin
      match argl with
      | [] -> (key :: accu, [])
      | arg :: argl -> (arg :: key :: accu, argl)
    end
  | Arg.Tuple specs ->
      Ulist.fold_left (append_args key) (accu, argl) specs
  | Arg.Rest _ -> (Ulist.rev_append argl (key :: accu), [])
;;

let parse_argv_only speclist_only argv =
  let keys_only = get_keys speclist_only in
  let (cmd, argl) =
    match Array.to_list argv with
    | [] -> co_invalid_arg "parse_argv_only" "empty vector of arguments"
    | cmd :: l -> (cmd, l) in
  let rec loop (to_parse, remains) = function
    | [] -> (Ulist.rev to_parse, Ulist.rev remains)
    | key :: argl ->
        if not (Ulist.mem key keys_only) then
          loop (to_parse, key :: remains) argl else
        let spec = find_spec key speclist_only in
        let (to_parse, argl) = append_args key (to_parse, argl) spec in
        loop (to_parse, remains) argl in
  let (to_parse, remains) = loop ([ cmd; ], [ cmd; ]) argl in
  Arg.current := 0;
  Arg.parse_argv
    (Array.of_list to_parse)
    speclist_only
    ignore
    (get_usage_msg ());
  Array.of_list remains
;;

let get_all_phase_command_options phase_name =
  try Hashtbl.find phase_command_options phase_name with
  | Not_found ->
      let Phase s = phase_name in
      co_failwith "get_all_phase_command_options"
        (Printf.sprintf "unknown phase %S" s)
;;

let get_all_phase phase_name =
  let phase_adv_cmd_opts = get_all_phase_command_options phase_name in
  let phase_cmd_opts =
    Ulist.flat_map expand_command_option phase_adv_cmd_opts in
  Ulist.map triple_of_regular_command_option phase_cmd_opts
;;

let parse_phase_argv phase_name argv =
  let phase_specs = get_all_phase phase_name in
  try parse_argv_only phase_specs argv with
  | Arg.Bad msg -> Printf.eprintf "%s" msg; exit 2;
  | Arg.Help msg -> Printf.printf "%s" msg; exit 0;
;;

let parse_phase phase_name =
  let argv = get_argv () in
  let argv = parse_phase_argv phase_name argv in
  set_argv argv;
;;
