(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2004-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: ulist.ml,v 1.28 2014/01/17 19:19:30 fclement Exp $ *)

(* More list operations. *)

let l_invalid_arg = Upervasives.invalid_arg "Ulist";;
let l_invalid_arg_no_msg fun_name = l_invalid_arg fun_name "";;

include List;;

type 'a listpn = 'a list;;
type 'a listn = 'a list;;
type 'a t = 'a listn;;

let min_length = Unatural._0_;;

let to_listpn = Upervasives.id;;
let of_listpn = Upervasives.id;;
let to_list = Upervasives.id;;
let of_list = Upervasives.id;;
let lift = Upervasives.id;;
let lift2 = Upervasives.id;;

let cons x l = x :: l;;

let length l = Unatural.unsafe_of_int (List.length l);;

let count fun_name n l =
  if n < 0 then l_invalid_arg fun_name "negative count" else
  let rec loop n accu l =
    if n = 0 then accu else
    match l with
    | [] -> accu
    | x :: rest -> loop (n - 1) (x :: accu) rest in
  loop n [] l
;;

let first n l = rev (count "first" n l);;

let last n l = count "last" n (rev l);;

let rec rank e = function
  | [] -> raise Not_found
  | x :: l -> if x = e then 0 else 1 + rank e l
;;

let gen_init stop next start f =
  let rec loop acc i =
    if stop i then acc else
    loop (f i :: acc) (next i)
  in
  loop [] start
;;

let init n f =
  let pn = pred (Unatural.to_int n) in
  gen_init (fun i -> i < 0) pred pn f
;;

let rev_init n f =
  let pn = pred (Unatural.to_int n) in
  gen_init (fun i -> i > pn) succ 0 f;;

let replace_by_list i l1 l2 =
  let rec loop i2 = function
    | [] -> []
    | x :: l ->
      if i2 <> i then x :: loop (i2 + 1) l else
      l1 @ loop (i2 + 1) l in
  loop 0 l2
;;

let rev_replace_by_list i l1 l2 =
  let rec loop accu i2 = function
    | [] -> accu
    | x :: l ->
      loop
        (if i2 <> i then x :: accu else rev_append l1 accu)
        (i2 + 1) l in
  loop [] 0 l2
;;

let insert i l1 l2 =
  let rec loop i2 = function
    | [] -> []
    | x :: l ->
      if i2 <> i then x :: loop (i2 + 1) l else
      (l1 @ [x]) @ loop (i2 + 1) l in
  loop 0 l2
;;

let rev_insert i l1 l2 =
  let rec loop accu i2 = function
    | [] -> accu
    | x :: l ->
      loop
        (if i2 <> i then x :: accu else x :: rev_append l1 accu)
        (i2 + 1) l in
  loop [] 0 l2
;;

(* Iterators. *)

let rec iter_i_shift f i = function
  | [] -> ()
  | x :: l -> f i x; iter_i_shift f (succ i) l
;;

let iter_i f = iter_i_shift f 0;;

let rec map_i_shift f i = function
  | [] -> []
  | x :: l -> let r = f i x in r :: map_i_shift f (succ i) l
;;

let map_i f = map_i_shift f 0;;

let rec rev_map_i_end_shift accu f i = function
  | [] -> accu
  | x :: l -> rev_map_i_end_shift (f i x :: accu) f (succ i) l
;;

let rev_map_i f = rev_map_i_end_shift [] f 0;;

let replace i0 x0 l = map_i (fun i x -> if i <> i0 then x else x0) l;;

let rev_replace i0 x0 l = rev_map_i (fun i x -> if i <> i0 then x else x0) l;;

let rev_flat_map_end accu f l =
  rev_append accu (fold_left (fun ll l -> rev_append (f l) ll) [] l)
;;

let rev_flat_map f = rev_flat_map_end [] f;;

let rec fold_left_i_shift f i accu = function
  | [] -> accu
  | x :: l -> let accu = f i accu x in fold_left_i_shift f (succ i) accu l
;;

let fold_left_i f = fold_left_i_shift f 0;;

let rev_flat_map_i_end_shift accu f i l =
  rev_append
    accu
    (fold_left_i_shift (fun j ll l -> rev_append (f j l) ll) i [] l)
;;

let rev_flat_map_i f = rev_flat_map_i_end_shift [] f 0;;

let flat_map_end accu f l = rev (rev_flat_map_end accu f l);;

let flat_map f = flat_map_end [] f;;

let flat_map_i_end_shift accu f i l =
  rev (rev_flat_map_i_end_shift accu f i l)
;;

let flat_map_i f = flat_map_i_end_shift [] f 0;;

let concat ll = flat_map Upervasives.id ll;;

let flatten = concat;;

let rec fold_right_i_shift f i l accu =
  match l with
  | [] -> accu
  | x :: l ->
    let accu = f i x accu in
    fold_right_i_shift f (succ i) l accu
;;

let fold_right_i f = fold_right_i_shift f 0;;

let rev_fold_left_env = fold_left;;

let fold_left_env f (env, accu) l =
  let (env, accu) = rev_fold_left_env f (env, rev accu) l in
  (env, rev accu)
;;

let fold_left_env_init f env l = fold_left_env f (env, []) l;;

(* Iterators on two lists. *)

let rec iter2_i_shift f i l1 l2 =
  match (l1, l2) with
  | ([], []) -> ()
  | (x1 :: l1, x2 :: l2) -> f i x1 x2; iter2_i_shift f (succ i) l1 l2
  | (_, _) -> l_invalid_arg_no_msg "iter2_i_shift"
;;

let iter2_i f = iter2_i_shift f 0;;

let rec map2_i_shift f i l1 l2 =
  match (l1, l2) with
  | ([], []) -> []
  | (x1 :: l1, x2 :: l2) ->
    let r = f i x1 x2 in
    r :: map2_i_shift f (succ i) l1 l2
  | (_, _) -> l_invalid_arg_no_msg "map2_i_shift"
;;

let map2_i f = map2_i_shift f 0;;

let rec rev_map2_i_end_shift accu f i l1 l2 =
  match (l1, l2) with
  | ([], []) -> accu
  | (x1 :: l1, x2 :: l2) ->
    rev_map2_i_end_shift (f i x1 x2 :: accu) f (succ i) l1 l2
  | (_, _) -> l_invalid_arg_no_msg "rev_map2_i_end_shift"
;;

let rev_map2_i f = rev_map2_i_end_shift [] f 0;;

let rev_flat_map2_end accu f l1 l2 =
  rev_append
    accu
    (fold_left2 (fun ll l1 l2 -> rev_append (f l1 l2) ll) [] l1 l2)
;;

let rev_flat_map2 f = rev_flat_map2_end [] f;;

let rec fold_left2_i_shift f i accu l1 l2 =
  match (l1, l2) with
  | ([], []) -> accu
  | (x1 :: l1, x2 :: l2) ->
    let accu = f i accu x1 x2 in
    fold_left2_i_shift f (succ i) accu l1 l2
  | (_, _) -> l_invalid_arg_no_msg "fold_left2_i_shift"
;;

let fold_left2_i f = fold_left2_i_shift f 0;;

let rev_flat_map2_i_end_shift accu f i l1 l2 =
  rev_append
    accu
    (fold_left2_i_shift
       (fun j ll l1 l2 -> rev_append (f j l1 l2) ll)
       i [] l1 l2)
;;

let rev_flat_map2_i f = rev_flat_map2_i_end_shift [] f 0;;

let flat_map2_end accu f l1 l2 = rev (rev_flat_map2_end accu f l1 l2);;

let flat_map2 f = flat_map2_end [] f;;

let flat_map2_i_end_shift accu f i l1 l2 =
  rev (rev_flat_map2_i_end_shift accu f i l1 l2)
;;

let flat_map2_i f = flat_map2_i_end_shift [] f 0;;

let rec fold_right2_i_shift f i l1 l2 accu =
  match (l1, l2) with
  | ([], []) -> accu
  | (x1 :: l1, x2 :: l2) ->
    let accu = f i x1 x2 accu in
    fold_right2_i_shift f (succ i) l1 l2 accu
  | (_, _) -> l_invalid_arg_no_msg "fold_right2_i_shift"
;;

let fold_right2_i f = fold_right2_i_shift f 0;;

let rev_fold_left2_env = fold_left2;;

let fold_left2_env f (env, accu) l1 l2 =
  let (env, accu) = rev_fold_left2_env f (env, rev accu) l1 l2 in
  (env, rev accu)
;;

let fold_left2_env_init f env l1 l2 = fold_left2_env f (env, []) l1 l2;;

(* Association lists. *)

let fold_assoc fun_name f = function
  | [] -> l_invalid_arg_no_msg fun_name
  | x0 :: l -> fold_left f x0 l
;;

let find_fst_assoc fun_name p =
  fold_assoc fun_name
    (fun (y0, _ as x0) (y, _ as x) -> if p y0 y then x else x0)
;;

let compare_assoc cmp =
  find_fst_assoc "compare_assoc" (fun y0 y -> cmp y0 y > 0)
;;

let minus_pervasives_compare x y = - Pervasives.compare x y;;

let min_assoc l = compare_assoc Pervasives.compare l
and max_assoc l = compare_assoc minus_pervasives_compare l
;;

let compare_fst_assoc = compare_assoc
and min_fst_assoc = min_assoc
and max_fst_assoc = max_assoc
;;

let find_snd_assoc fun_name p =
  fold_assoc fun_name
    (fun (_, y0 as x0) (_, y as x) -> if p y0 y then x else x0)
;;

let compare_snd_assoc cmp =
  find_snd_assoc "compare_snd_assoc" (fun y0 y -> cmp y0 y > 0)
;;

let min_snd_assoc l = compare_snd_assoc Pervasives.compare l
and max_snd_assoc l = compare_snd_assoc minus_pervasives_compare l
;;

(* Sorting. *)

let of_range pred l r =
  let rec loop accu r =
    if r > l then loop (r :: accu) (pred r) else
    if r = l then l :: accu else accu in
  loop [] (pred r)
;;
