(***********************************************************************)
(*                                                                     *)
(*                              Ref-image                              *)
(*                                                                     *)
(*                  Image Segmentation by Refinement                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2009-2013 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mapping_types.mli,v 1.1 2013/06/13 19:01:17 fclement Exp $ *)

(** {3 Types for mappings} *)

(** {6 Extensional mappings} *)

(** Extensional mappings are simply (finite) functions (or mappings) defined
  as the (finite) set of the points in its graph. In other words, we
  represent the function f : A -> B, by its graph
  G = \{ (x, y) in A x B, y = f x \}. *)

type ('a, 'b) extensional_mapping = ('a, 'b) Ugraph.t;;
(** The graph of the mapping is simply represented by a set of arrows. *)

(** {6 Domains of mappings} *)

type 'a domain = 'a Uset.t;;
(** A domain is the definitional set of a mapping, or the set of elements
  that have an image by the mapping.
  The domain of f : A -> B is the set D such that:
  D = \{ x in A, exists y in B, y = f x \}

  A domain is given either as a set of elements (extensional view of the
  set), or using a ``belongs-to'' predicate (for the intensional view of the
  set).

  Note that intensional views are more powerful (and very often more compact)
  since an intensional view may easily specify an infinite set. For instance,
  [Intensional (fun x -> x > 0)] is the set of positive integer numbers. *)

(** {6 Intensional mappings} *)

(** Intensional mappings are functions given by a formula and a domain for
   which the formula applies. *)

type ('a, 'b) intensional_mapping =
  {
   domain : 'a domain;
   application : 'a Uset.uindex -> 'b;
  }
;;
(** An intensional mapping is given by an application (a total function on its
  domain) and a definitional domain.

  The condition of totality of the formula on the domain cannot be checked
  statically: the user should not create an intensional mapping with a
  partial function that is not total on its domain! *)

(** {6 Structural mappings} *)

(** A structural mapping is a function defined with multiple partial
  functions defined on subsets of the domain of the structural mapping.
  The intuitive idea is the notion of patch: given a function f defined on
  some domain F, then the patch of f by the function g defined on a subdomain
  G of F is the function h, defined on F as:
  for all x in F,
  if x belongs to G then h (x) = g (x), otherwise h (x) = f (x). *)

type ('a, 'b) intensional_structural_mapping =
  ('a, 'b) intensional_mapping Patchwork_types.t
;;
(** A structural mapping is a patchwork of intensional mappings. *)

type ('a, 'b) intensional_mapping_location =
  ('a, 'b) intensional_mapping Patchwork_types.location
;;

type ('a, 'b) extensional_structural_mapping =
  ('a, 'b) extensional_mapping Ulist1.list1
;;
(** A structural mapping is a patchwork of extensional mappings. *)

type ('a, 'b) extensional_mapping_location =
  ('a, 'b) extensional_mapping Ulist1_zipper.location
;;
